"""Module of logging"""
from typing import ClassVar, List, Optional, TypeVar, Union
import abc
import functools
import os
import sys
import warnings
import logging
import logging.config
from logging import LogRecord
import datetime
import multiprocessing
import time
import rich.logging
from rich.console import ConsoleRenderable
from rich.text import Text

from vlutils.base.decoratorContextManager import DecoratorContextManager
from .io import rotateItems
from .runtime import functionFullName


__all__ = [
    "WaitingBar",
    "LoggingDisabler",
    "configLogging",
    "readableSize"
]

T = TypeVar("T")

class LoggerBase(abc.ABC):
    @abc.abstractmethod
    def setLevel(self, *_, **__):
        raise NotImplementedError

    @abc.abstractmethod
    def debug(self, *_, **__):
        raise NotImplementedError

    @abc.abstractmethod
    def info(self, *_, **__):
        raise NotImplementedError

    @abc.abstractmethod
    def warning(self, *_, **__):
        raise NotImplementedError

    @abc.abstractmethod
    def warn(self, *_, **__):
        raise NotImplementedError

    @abc.abstractmethod
    def error(self, *_, **__):
        raise NotImplementedError

    @abc.abstractmethod
    def exception(self, *_, **__):
        raise NotImplementedError

    @abc.abstractmethod
    def critical(self, *_, **__):
        raise NotImplementedError

    @abc.abstractmethod
    def log(self, *_, **__):
        raise NotImplementedError


def trackingFunctionCalls(function: T, logger: Union[logging.Logger, LoggerBase] = logging.root) -> T:
    fullName = functionFullName(function)
    if isinstance(function, functools.partial):
        funcArgs = function.args
        funcKwArgs = function.keywords
    else:
        funcArgs = ()
        funcKwArgs = dict()
    def wrapper(*args, **kwArgs):
        allArgs = ", ".join(str(arg) for arg in (args + funcArgs))
        # Python 3.9+
        allkwArgs = ", ".join(f"{key}={value}" for key, value in (kwArgs | funcKwArgs).items())
        if len(allArgs) > 0:
            logger.debug("Call %s(%s, %s)", fullName, allArgs, allkwArgs)
        else:
            logger.debug("Call %s(%s)", fullName, allkwArgs)
        return function(*args, **kwArgs)
    return wrapper


def readableSize(byteSize: int, floating: int = 2, binary: bool = True) -> str:
    """Convert bytes to human-readable string (like `-h` option in POSIX).

    Args:
        size (int): Total bytes.
        floating (int, optional): Floating point length. Defaults to 2.
        binary (bool, optional): Format as XB or XiB. Defaults to True.

    Returns:
        str: Human-readable string of size.
    """
    size = float(byteSize)
    unit = "B"
    if binary:
        for unit in ["B", "kiB", "MiB", "GiB", "TiB", "PiB"]:
            if size < 1024.0 or unit == "PiB":
                break
            size /= 1024.0
        return f"{size:.{floating}f}{unit}"
    for unit in ["B", "kB", "MB", "GB", "TB", "PB"]:
        if size < 1000.0 or unit == "PB":
            break
        size /= 1000.0
    return f"{size:.{floating}f}{unit}"


class WaitingBar(DecoratorContextManager):
    """A CLI tool for printing waiting bar.

    Example:
    ```python
        @WaitingBar("msg")
        def longTime():
            # Long time operation
            ...

        with WaitingBar("msg"):
            # Long time operation
            ...
    ```

    Args:
        msg (str): Addtional message shows after bar.
        ncols (int): Total columns of bar.
    """
    def __init__(self, msg: str, ncols: int = 10):
        if ncols <= 8:
            raise ValueError("ncols must greater than 8, got %d", ncols)
        self._msg = msg
        self._ticker = None
        self._stillRunning = None
        self._ncols = ncols
        self.animation = list()
        # "       =       "
        template = (" " * (ncols + 1) + "=" * (ncols - 8) + " " * (ncols + 1))
        for i in range(2 * (ncols - 2)):
            start = 2 * (ncols - 2) - i
            end = 3 * (ncols - 2) - i
            self.animation.append("[" + template[start:end] + "]" + r" %s")

    def __enter__(self):
        self._stillRunning = multiprocessing.Value("b", True)
        self._ticker = multiprocessing.Process(name="waitingBarTicker", target=self._print, args=[self._stillRunning])
        self._ticker.start()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._stillRunning.value = False
        self._ticker.join()
        print(" " * (len(self._msg) + self._ncols + 1), end="\r", file=sys.stderr)

    def _print(self, stillRunning: multiprocessing.Value):
        i = 0
        while bool(stillRunning.value):
            print(self.animation[i % len(self.animation)] % self._msg, end='\r', file=sys.stderr)
            time.sleep(.06)
            i += 1


class LoggingDisabler:
    """Disable or enable logging temporarily.

    Example:
    ```python
        # True -> disable logging, False -> enable logging
        with LoggingDisabler(logger, True):
            # Some operations
            ...
    ```

    Args:
        logger (logging.Logger): The target logger to interpolate.
        disable (bool): Whether to disable logging.
    """
    def __init__(self, logger: logging.Logger, disable: bool):
        self._logger = logger
        self._disable = disable
        self._previous_status = False

    def __enter__(self):
        if self._disable:
            self._previous_status = self._logger.disabled
            self._logger.disabled = True

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._disable:
            self._logger.disabled = self._previous_status


class _DeprecationFilter:
    def filter(self, record: LogRecord):
        if "depreca" in record.msg:
            return 0
        return 1


class KeywordRichHandler(rich.logging.RichHandler):
    KEYWORDS: ClassVar[Optional[List[str]]] = [
        r"(?P<green>\b([gG]ood|[bB]etter|[bB]est|[sS]uccess(|ful|fully))\b)",
        r"(?P<magenta>\b([bB]ase|[cC]all(|s|ed|ing)|[Mm]ount(|s|ed|ing))\b)",
        r"(?P<cyan>\b([mM]aster|nccl|NCCL|[mM]ain|···|[tT]otal|[tT]rain(|s|ed|ing)|[vV]alidate(|s|d)|[vV]alidat(|ing|ion)|[tT]est(|s|ed|ing))\b)",
        r"(?P<yellow>\b([lL]atest|[lL]ast|[sS]tart(|s|ed|ing)|[bB]egin(|s|ning)|[bB]egun|[cC]reate(|s|d|ing)|[gG]et(|s|ting)|[gG]ot|)\b)",
        r"(?P<red>\b([eE]nd(|s|ed|ing)|[fF]inish(|es|ed|ing)|[kK]ill(|s|ed|ing)|[iI]terrupt(|s|ed|ting)|[qQ]uit|QUIT|[eE]xit|EXIT|[bB]ad|[wW]orse|[sS]low(|er))\b)"
    ]

    def render_message(self, record: LogRecord, message: str) -> ConsoleRenderable:
        use_markup = getattr(record, "markup", self.markup)
        message_text = Text.from_markup(message) if use_markup else Text(message)

        highlighter = getattr(record, "highlighter", self.highlighter)

        if self.KEYWORDS:
            for keyword in self.KEYWORDS:
                message_text.highlight_regex(keyword)
                # message_text.highlight_words(value, key, case_sensitive=False)

        if highlighter:
            message_text = highlighter(message_text)

        return message_text


def configLogging(logDir: str, rootName: str = "", level: Union[str, int] = logging.INFO, logName: str = None, rotateLogs: int = 10, ignoreWarnings: list = None) -> logging.Logger:
    """Logger configuration.

    Args:
        logDir (str): Log files placed in this folder.
        rootName (str): Logger's root name.
        level (str): Minimal level that will be logged.
        useTqdm (bool, optional): Not used. Defaults to False.
        logName (str, optional): Log file name, if None, use the formatted current time. Defaults to None.
        rotateLogs (int, optional): Whether to perform rotate in this folder, -1 is don't rotate. Defaults to 10.
        ignoreWarnings (list, optional): Which warnings don't want to be logged. Defaults to None.

    Returns:
        logging.Logger: The configed logger.
    """
    os.makedirs(logDir, exist_ok=True)
    if rotateLogs > 0:
        rotateItems(logDir, rotateLogs)
    if logName is None:
        fPrefix = os.path.join(logDir, "{0}".format(datetime.datetime.now().strftime(r"%y%m%d-%H%M%S")))
    else:
        fPrefix = os.path.join(logDir, logName)
    logging_config = {
        "version": 1,
        "formatters": {
            "full": {
                "format": r"%(asctime)s - %(name)s - %(levelname)s - %(message)s"
            },
            "simple": {
                "format": r"%(asctime)s - %(message)s",
                "datefmt": r"%m/%d %H:%M:%S"
            }
        },
        "filters": {
            "deprecation": {
                "()": _DeprecationFilter
            }
        },
        "handlers": {
            "console": {
                "class": "vlutils.logger.KeywordRichHandler",
                "level": level,
                "rich_tracebacks": True,
                "tracebacks_show_locals": True,
                "log_time_format": r"%m/%d %H:%M"
            },
            "info_file": {
                "class": "logging.FileHandler",
                "level": level,
                "formatter": "full",
                "filename": f"{fPrefix}.log",
                "mode": "w"
            }
        },
        "loggers": {
            rootName: {
                "propagate": False,
                "level": level,
                "handlers": [
                    "console",
                    "info_file"
                ]
            }
        }
    }
    logging.config.dictConfig(logging_config)

    def handleException(exc_type, exc_value, exc_traceback):
        if issubclass(exc_type, KeyboardInterrupt):
            sys.__excepthook__(exc_type, exc_value, exc_traceback)
            return
        logger = logging.getLogger(rootName)
        logger.exception(repr(exc_value), exc_info=(exc_type, exc_value, exc_traceback))
    sys.excepthook = handleException

    def handleWarning(message, category, filename, lineno, file=None, line=None):
        _ = file
        logger = logging.getLogger(rootName)
        if ignoreWarnings is not None and category in ignoreWarnings:
            return
        logger.warning(warnings.formatwarning(message, category, filename, lineno, line))
    warnings.showwarning = handleWarning
    return logging.getLogger(rootName)



"""Module of runtime utils"""
import functools
from typing import Callable, Union
from types import FunctionType, MethodType
import os
import logging
from pprint import pformat
import time
from typing import List, Dict, Tuple

from .types import StrPath

import torch
import pynvml


__all__ = [
    "preAllocateMem",
    "queryGPU",
    "gpuInfo",
    "Timer"
]

def inspectFunction(function: Callable) -> Callable:
    if isinstance(function, functools.partial):
        return function.func
    return function

def _fullName(function):
    if hasattr(function, "__qualname__"):
        return f"{function.__module__}.{function.__qualname__}"
    if hasattr(function, "__class__"):
        return f"{function.__module__}.{function.__class__.__name__}"
    return repr(function)

def functionFullName(function: Callable) -> str:
    if isinstance(function, functools.partial):
        return functionFullName(function.func)
    if isinstance(function, (FunctionType, MethodType)):
        return _fullName(function)
    if (function.__str__ is not object.__str__) or (type(function).__str__ is not object.__str__):
        return function.__str__()
    return _fullName(function)

def relativePath(path: StrPath, start: Union[StrPath, None] = None):
    result = os.path.relpath(path, start)
    if not result.startswith("../"):
        result = "./" + result
    return result

def preAllocateMem(memSize: int):
    """Pre-allocate VRAM in GPUs.

    Args:
        memSize (int): Preserved VRAM amount (MiB).
    """
    devices = torch.cuda.device_count()
    for d in range(devices):
        x = torch.rand((256, 1024, memSize), device=f"cuda:{d}")
        del x
    return


def gpuInfo() -> List[Dict[str, int]]:
    """Helper for list all gpus.

    Returns:
        List[Dict[str, int]]: A list of dicts { "memory.used": int, "memory.total": int }, sorted by `CUDA_DEVICE_ORDER`
    """
    pynvml.nvmlInit()
    deviceCount = pynvml.nvmlDeviceGetCount()
    gpus = list()
    for i in range(deviceCount):
        handle = pynvml.nvmlDeviceGetHandleByIndex(i)
        info = pynvml.nvmlDeviceGetMemoryInfo(handle)
        gpus.append({"memory.used": info.used / 1048576, "memory.total": info.total / 1048576})
    return gpus


def queryGPU(wantsMore: bool = False, givenList: list = None, needGPUs: int = -1, needVRamEachGPU: int = -1, writeOSEnv: bool = True, logger: logging.Logger = None) -> List[Tuple[int, int]]:
    """Query GPUs that meet requirements.

    Example:
    ```python
        # Find a GPU has free VRAM >= 2000MiB
        queryGPU(needGPUs=1, needVRamEachGPU=2000)
        # Only when all GPUs are free, otherwise raise EnvironmentError
        queryGPU(needGPUs=-1, needVRamEachGPU=-1)
        # Don't write CUDA_VISIBLE_DEVICE, return real GPU ID
        gpuList = queryGPU(writeOSEnv=False)
        # A typical usage
        while True:
            try:
                queryGPU(needGPUs=3, needVRamEachGPU=8000)
                # Go to run
                break
            except EnvironmentError:
                # Wait 15s
                time.sleep(1500)
            # Begin GPU tasks
            run()
    ```

    Args:
        wantsMore (bool, optional): Wants at least `needGPUs` GPUs, if there are more GPUs, use them all. Defaults to False.
        givenList (list, optional): If given, only GPUs id in this list will be queried. Defaults to None.
        needGPUs (int, optional): How many GPUs take in demand. Defaults to -1.
        needVRamEachGPU (int, optional): It is OK to use a GPU if free VRAM (MiB) is larger than this threshold. Defaults to -1.
        writeOSEnv (bool, optional): Overwrite env parameter 'CUDA_VISIBLE_DEVICES'. Defaults to True.
        logger (logging.Logger, optional): For logging. Defaults to None.

    Raises:
        EnvironmentError: No GPU satisfied requirements.

    Returns:
        List[Tuple[int, int]]: List of available gpu id and free VRAM.
    """
    logger = logger or logging

    # keep the devices order same as in nvidia-smi
    os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"

    gpus = gpuInfo()
    if needGPUs < 0:
        needGPUs = len(gpus)

    # logger.debug("\n" + str(gpus))
    if isinstance(givenList, list):
        it = givenList
        needGPUs = min(needGPUs, len(givenList))
    else:
        it = range(len(gpus))

    gpus = [(i, gpus[i]) for i in it]
    if wantsMore:
        gpus = sorted(gpus, key=lambda item: item[1]['memory.used'])

    gpuList = []
    for i, g in gpus:
        if needVRamEachGPU < 0:
            if g['memory.used'] < 64:
                # give space for basic vram
                gpuList.append((i, (g['memory.total'] - g['memory.used'] - 64)))
                logger.debug("adding gpu[%d] with %.2fMB free.", i, g['memory.total'] - g['memory.used'])
        elif g['memory.total'] - g['memory.used'] > needVRamEachGPU + 64:
            gpuList.append((i, (g['memory.total'] - g['memory.used'] - 64)))
            logger.debug("adding gpu[%d] with %.2fMB free.", i, g['memory.total'] - g['memory.used'])
        if len(gpuList) >= needGPUs and not wantsMore:
            break

    if len(gpuList) >= needGPUs:
        # keep order
        gpuList = sorted(gpuList, key=lambda item: item[0])
        logger.debug("Found %d %s satisfied with args %s.", len(gpuList), "gpu" if len(gpuList) == 1 else "gpus", pformat(
            {
                "wantsMore": wantsMore,
                "givenList": givenList,
                "needGPUs": needGPUs,
                "needVRamEachGPU": needVRamEachGPU
            }, indent=4
        ))
        if writeOSEnv:
            os.environ["CUDA_VISIBLE_DEVICES"] = ",".join(map(str, [item[0] for item in gpuList]))
            newGPUList = []
            j = 0
            for i, mem in gpuList:
                newGPUList.append((j, mem))
                j += 1
            gpuList = newGPUList
        else:
            try:
                os.environ.pop("CUDA_VISIBLE_DEVICES")
            except KeyError:
                pass
        return gpuList
    else:
        raise EnvironmentError("Current system status is not satisfied")


class Timer:
    """A simple timer

    Example:
    ```python
        # Start the timer
        timer = Timer()
        ...
        # Last interval and total spent from start
        # 3.0,    3.0
        interval, total = timer.Tick()
        ...
        # 4.4,    7.4
        interval, total = timer.Tick()
    ```
    """
    def __init__(self):
        self._initial = self._tick = time.time()

    def tick(self):
        tock = time.time()
        interval = tock - self._tick
        self._tick = tock
        return interval, self._tick - self._initial



"""Module of Restorable class"""
from typing import Any, Dict
import abc


__all__ = [
    "Dictable",
    "Restorable"
]


class Dictable(abc.ABC):
    """An abstract class implements PyTorch-like state-dict."""

    @abc.abstractmethod
    def state_dict(self) -> Dict[str, Any]:
        """Export current state-dict

        Returns:
            Dict[str, Any]: The state-dict.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def load_state_dict(self, stateDict: Dict[str, Any]):
        """Load from state-dict

        Args:
            stateDict (Dict[str, Any]): State-dict to be loaded.
        """
        raise NotImplementedError


class Restorable(Dictable):
    """A class implements PyTorch-like state-dict.

    Example:
    ```python
        class Foo(Restorable):
            super().__init__()
            # all fields start with '_' will be stored in the state-dict.
            self._x = 3

        foo = Foo()
        # { "_x": 3 }
        foo.state_dict()
        foo.load_state_dict({ "_x": 1 })
    ```

    Attributes:
        valuesToSave (set): names of all attributes to be saved.
    """
    def __init__(self):
        self.valuesToSave = set()

    def state_dict(self):
        return {key: self.__dict__[key].state_dict() if callable(getattr(self.__dict__[key], "state_dict", None)) else self.__dict__[key] for key in self.valuesToSave}

    def load_state_dict(self, stateDict: Dict[str, Any], strict: bool = True):
        for key, value in self.__dict__.items():
            if not key.startswith("_"):
                continue
            if callable(getattr(value, "load_state_dict", None)):
                value.load_state_dict(stateDict[key])  # pylint: disable = no-member
            elif key not in stateDict:
                if strict:
                    raise RuntimeError(f"{key} not in stateDict.")
                else:
                    continue
            else:
                self.__dict__[key] = stateDict[key]

    def __setattr__(self, name: str, value: Any):
        self.__dict__[name] = value
        if hasattr(self, "valuesToSave"):
            if name.startswith("_"):
                self.valuesToSave.add(name)



"""Module of DecoratorContextManager.

Originally from https://github.com/pytorch/pytorch/blob/671ee71ad4b6f507218d1cad278a8e743780b716/torch/autograd/grad_mode.py#L16
"""
import inspect
import functools
import abc


__all__ = [
    "DecoratorContextManager"
]


class DecoratorContextManager(abc.ABC):
    """Allow a context manager to be used as a decorator

    Example:
    ```python
        class Foo(DecoratorContextManager):
            ...

            def __enter__(self):
                ...

            def __exit__(self, exc_type, exc_val, exc_tb):
                ...

        def add(x, y):
            return a + b
        @Foo()
        def addD(x, y):
            return a + b

        # Equivalent
        with Foo():
            add(3, 4)
        # Equivalent
        addD(3, 4)
    ```
    """

    def __call__(self, func):
        if inspect.isgeneratorfunction(func):
            return self._wrap_generator(func)

        @functools.wraps(func)
        def decorate_context(*args, **kwargs):
            with self:
                return func(*args, **kwargs)
        return decorate_context

    def _wrap_generator(self, func):
        """Wrap each generator invocation with the context manager"""
        @functools.wraps(func)
        def generator_context(*args, **kwargs):
            gen = func(*args, **kwargs)
            while True:
                try:
                    with self:
                        x = next(gen)
                    yield x
                except StopIteration:
                    break
        return generator_context

    @abc.abstractmethod
    def __enter__(self):
        raise NotImplementedError

    @abc.abstractmethod
    def __exit__(self, exc_type, exc_val, exc_tb):
        raise NotImplementedError



"""Package of Base classes.

Exports:
    DecoratorContextManager: Allow a context manager to be used as a decorator.
    Dictable: An interface with `state_dict()` and `load_state_dict()` method.
    Restorable: A class implements PyTorch-like state_dict.
    DataParallel: Modified `nn.DataParallel` runs class-method decorated by `parallelFunction` parallelly.
    parallelFunction: A decorator for class-methods that want to parallel-apply.
    Factory: Factory design mode.
    Module: Custom `torch.nn.Module` with function mapping.
"""
from .dataParallel import DataParallel, parallelFunction
from .decoratorContextManager import DecoratorContextManager
from .registry import Registry
from .restorable import Restorable, Dictable
from .module import Module
from .freqHook import FrequecyHook


__all__ = ["DataParallel",
           "parallelFunction",
           "Registry",
           "Restorable",
           "Dictable",
           "Module",
           "DecoratorContextManager",
           "FrequecyHook"]



"""Module of Saver"""
from typing import Any, Dict, Union, Optional
import os
import logging
import shutil
import datetime
from pathlib import Path

import torch
from tensorboard import program
from torch.utils.tensorboard import SummaryWriter
import yaml

from .io import rotateItems
from .config import serialize
from .logger import configLogging, LoggerBase
from .types import StrPath
from .runtime import relativePath

__all__ = [
    "Saver"
]


class Saver(SummaryWriter, LoggerBase):
    """A class for load and save model

    Example:
    ```python
        #  Save path: `saved/myModel`.
        #     config: Any object that can dump to yaml.
        # autoManage: If True, rename `latest` folder to current time and recreate `latest` folder.
        #   maxItems: Keep items (files/folders) in the save path no more than it. If <= 0, ignore.
        #    reserve: When autoManage is on and `latest` folder exists, continue to save file in current `latest` folder other than rotate it.
        #   dumpFile: Dump source code and config to the save path.
        saver = Saver("saved/myModel", config, autoManage=True, maxItems=10, reserve=False, dumpFile="mySoureCode/path")
    ```

    Args:
        saveDir (str): Direcotry to save model
        config (Any): The config that can dump to yaml
        autoManage (bool, optional): Auto create `latest` folder and rotate folders by time to save the new model. Defaults to True.
        maxItems (int, optional): Max old checkpoints to preserve. Defaults to 25.
        reserve (bool, optional): When autoManage is on and `latest` folder exists, continue to save file in current `latest` folder other than rotate it. Defaults to False.
        dumpFile (str, optional): The path of any folder want to save.
    """
    NewestDir = "latest"

    @staticmethod
    def composePath(saveDir: StrPath, saveName: StrPath = "saved.ckpt", autoManage: bool = True):
        if saveDir.endswith(Saver.NewestDir):
            autoManage = False
        if autoManage:
            _saveDir = os.path.join(saveDir, Saver.NewestDir)
        else:
            _saveDir = saveDir
        return os.path.join(_saveDir, saveName)

    def __init__(self, saveDir: StrPath, saveName: StrPath = "saved.ckpt", loggerName: str = "root", loggingLevel: str = "INFO", config: Optional[Any] = None, autoManage: bool = True, maxItems: int = 25, reserve: Optional[bool] = False, dumpFile: Optional[str] = None, activateTensorboard: bool = True):
        if saveDir.endswith(self.NewestDir):
            autoManage = False

        if autoManage:
            if os.path.exists(os.path.join(saveDir, self.NewestDir)) and not reserve:
                newDir = os.path.join(saveDir, datetime.datetime.now().strftime(r"%y%m%d-%H%M%S"))
                shutil.move(os.path.join(saveDir, self.NewestDir), newDir)
                # self.debug("Auto rename %s to %s", os.path.join(saveDir, self.NewestDir), newDir)
            os.makedirs(os.path.join(saveDir, self.NewestDir), exist_ok=True)
            if maxItems > 0:
                rotateItems(saveDir, maxItems)
            self._saveDir = os.path.join(saveDir, self.NewestDir)
        else:
            self._saveDir = saveDir
        super().__init__(self._saveDir)

        logger = configLogging(self.SaveDir, loggerName, loggingLevel, rotateLogs=-1)
        self.Logger = logger

        self._savePath = os.path.join(self._saveDir, saveName)
        self.debug("Saver located at %s", relativePath(self._saveDir))


        if config is not None:
            with open(os.path.join(self._saveDir, "config.yaml"), "w") as fp:
                yaml.dump(serialize(config), fp)
        if dumpFile is not None and not str.isspace(dumpFile) and os.path.exists(dumpFile):
            self._dumpFile(dumpFile)

        self._infoCounter = 0

        if activateTensorboard:
            tb = program.TensorBoard()
            tb.configure(argv=[None, "--logdir", self._saveDir, "--load_fast", "false"])
            url = tb.launch()
            self.debug("Tensorboard is listening at %s", url)
            self._url = url
        else:
            self._url = None

    @property
    def TensorboardURL(self) -> str:
        if self._url is not None:
            return self._url
        else:
            raise RuntimeError("No tensorboard listening.")

    def setLevel(self, level):
        pass

    def debug(self, msg, *args, **kwargs):
        pass

    def info(self, msg, *args, **kwargs):
        pass

    def warning(self, msg, *args, **kwargs):
        pass

    def warn(self, msg, *args, **kwargs):
        pass

    def error(self, msg, *args, **kwargs):
        pass

    def exception(self, msg, *args, exc_info=True, **kwargs):
        pass

    def critical(self, msg, *args, **kwargs):
        pass

    def log(self, level, msg, *args, **kwargs):
        pass

    @property
    def Logger(self) -> logging.Logger:
        return self._logger

    @Logger.setter
    def Logger(self, logger: logging.Logger):
        logger = logger or logging
        self._logger = logger
        self.setLevel = self._logger.setLevel
        self.debug = self._logger.debug
        self.info = self._logger.info
        self.warning = self._logger.warning
        self.warn = self._logger.warn
        self.error = self._logger.error
        self.exception = self._logger.exception
        self.critical = self._logger.critical
        self.log = self._logger.log

    def _dumpFile(self, path: StrPath):
        shutil.copytree(path, os.path.join(self._saveDir, "dump"), symlinks=True, ignore=lambda src, path: [x for x in path if x == "__pycache__"], ignore_dangling_symlinks=True)

    @property
    def SaveDir(self) -> Path:
        """Return the current saving directory path"""
        return Path(self._saveDir)

    @property
    def SavePath(self) -> Path:
        """Return the current saving ckpt absolute path"""
        return Path(self._savePath)

    def moveTo(self, dest: StrPath):
        """Move current saving dir to dest

        Args:
            dest (str): The path of destination dir, execute by shutil.move().
        """
        self.flush()
        self.close()
        shutil.move(self._saveDir, dest)
        self._saveDir = dest
        self._savePath = os.path.join(dest, self.SavePath.name)
        self.log_dir = self._saveDir

    def save(self, path: StrPath = None, **objs: Any):
        """Save anything

        Args:
            **objs (Any): The saved items ordered by names.
        """
        saveDict = dict()
        for key, value in objs.items():
            if isinstance(value, torch.nn.Module):
                saveDict[key] = value.state_dict()
            elif hasattr(value, "state_dict"):
                saveDict[key] = value.state_dict()
            else:
                saveDict[key] = value
        torch.save(saveDict, path or self._savePath)
        self.debug("Successfully saved checkpoint with keys: %s", list(saveDict.keys()))

    @staticmethod
    def load(filePath: StrPath, mapLocation: Dict[str, str] = None, strict: bool = True, logger: Optional[Union[logging.Logger, "Saver"]] = None, **objs: Any) -> Dict[str, Any]:
        """Load from ckpt.

        Args:
            filePath (str): The destination path to load ckpt.
            mapLocation (Dict[str, str], optional): See torch.load(mapLocation). Defaults to None.
            strict (bool, optional): See torch.load_state_dict(strict). Defaults to True.
            logger (Logger, optional): For logging. Defaults to None.
            **objs (Any): Anything to load by name. If it has `.load_state_dict()`, use this method. Else the loaded item will be placed in resulting dict.

        Returns:
            Dict[str, Any]: The loaded dict.
        """
        savedDict = torch.load(filePath, map_location=mapLocation)
        logger = logger or logging
        logger.debug("Load state_dict with keys:\r\n%s", savedDict.keys())
        for key, value in objs.items():
            stateDict = savedDict[key]
            if isinstance(value, torch.nn.Module):
                value.load_state_dict(stateDict, strict=strict)
            elif callable(getattr(value, "load_state_dict", None)):
                value.load_state_dict(stateDict)
            else:
                if isinstance(value, torch.Tensor):
                    value.data = stateDict
                else:
                    objs[key] = stateDict
        return objs


class DummySaver(Saver, LoggerBase):
    def __init__(self, saveDir: StrPath, saveName: StrPath = "saved.ckpt", loggerName: str = "root", loggingLevel: str = "INFO", config: Optional[Any] = None, autoManage: bool = True, maxItems: int = 25, reserve: Optional[bool] = False, dumpFile: Optional[str] = None, activateTensorboard: bool = False):
        if saveDir.endswith(self.NewestDir):
            autoManage = False
        if autoManage:
            self._saveDir = os.path.join(saveDir, self.NewestDir)
        else:
            self._saveDir = saveDir
        self._savePath = os.path.join(self._saveDir, saveName)

    @property
    def Logger(self) -> logging.Logger:
        raise NotImplementedError("Dummy saver does not have logger.")

    @Logger.setter
    def Logger(self, logger: logging.Logger):
        raise NotImplementedError("Dummy saver does not have logger.")

    def setLevel(self, level):
        pass

    def debug(self, msg, *args, **kwargs):
        pass

    def info(self, msg, *args, **kwargs):
        pass

    def warning(self, msg, *args, **kwargs):
        pass

    def warn(self, msg, *args, **kwargs):
        pass

    def error(self, msg, *args, **kwargs):
        pass

    def exception(self, msg, *args, exc_info=True, **kwargs):
        pass

    def critical(self, msg, *args, **kwargs):
        pass

    def log(self, level, msg, *args, **kwargs):
        pass

    def _dumpFile(self, path: StrPath):
        pass

    def moveTo(self, dest: StrPath):
        raise NotImplementedError("Dummy saver does not implement `moveTo` function.")

    def save(self, path: StrPath = None, **objs: Any):
        raise NotImplementedError("Dummy saver does not implement `save` function.")

    @staticmethod
    def load(filePath: StrPath, mapLocation: Dict[str, str] = None, strict: bool = True, logger: logging.Logger = None, **objs: Any) -> Dict[str, Any]:
        savedDict = torch.load(filePath, map_location=mapLocation)
        for key, value in objs.items():
            stateDict = savedDict[key]
            if isinstance(value, torch.nn.Module):
                value.load_state_dict(stateDict, strict=strict)
            elif callable(getattr(value, "load_state_dict", None)):
                value.load_state_dict(stateDict)
            else:
                if isinstance(value, torch.Tensor):
                    value.data = stateDict
                else:
                    objs[key] = stateDict
        return objs



"""Module for file operations"""
import logging

import os
import shutil
import atexit
import time
from subprocess import Popen, PIPE


__all__ = [
    "openTensorboard",
    "deleteDir",
    "deleteFilesOlderThan",
    "rotateItems"
]


def openTensorboard(log_dir: str, port: int):
    process = Popen(["tensorboard", "--logdir", log_dir, "--port", str(port)], stdout=PIPE, stderr=PIPE)

    def cleanup():
        process.terminate()
        timeout_sec = 5
        for _ in range(timeout_sec):
            if process.poll() is None:
                time.sleep(1)
            else:
                logging.info("Tensorboard quitted")
                return
        process.kill()
        logging.info("Tensorboard killed")
    atexit.register(cleanup)
    logging.info("Tensorboard opened at %d", port)


def deleteDir(path):
    shutil.rmtree(path, ignore_errors=True)


def deleteFilesOlderThan(folder: str, seconds: int):
    now = time.time()
    for f in os.listdir(folder):
        if os.path.isfile(os.path.join(folder, f)):
            if os.stat(os.path.join(folder, f)).st_mtime < now - seconds:
                os.remove(os.path.join(folder, f))


def rotateItems(folder: str, count: int):
    """Rotate items in a directory (Old item will be delete until total-amount < desired).

    Args:
        folder (str): Directory to rotate.
        count (int): How many newer items want to preserve.
    """
    file_list = []
    for f in os.listdir(folder):
        file_list.append(f)
    file_list = sorted(file_list, key=lambda f: os.stat(os.path.join(folder, f)).st_mtime)
    if len(file_list) <= count:
        return
    for f in file_list[:-count]:
        f = os.path.join(folder, f)
        if os.path.isfile(f):
            os.remove(f)
        else:
            shutil.rmtree(f)



from typing import Any

import torch
import numpy as np


__all__ = [
    "recursiveCompare"
]


def recursiveCompare(a: Any, b: Any) -> bool:
    if isinstance(a, (list, set, tuple)):
        return isinstance(b, (list, set, tuple)) and all(recursiveCompare(x, y) for x, y in zip(a, b))
    elif isinstance(a, dict):
        if not isinstance(b, dict):
            return False
        for key, value in a.items():
            if key not in b:
                return False
            if not recursiveCompare(value, b[key]):
                return False
        return True
    else:
        if type(a) != type(b):
            return False
        if isinstance(a, torch.Tensor):
            return torch.equal(a, b)
        if isinstance(a, np.ndarray):
            return np.array_equal(a, b)
        return a == b



import numpy as np
import torch
# import pytest
from skimage.color import rgb2hsv as sk_rgb2hsv, hsv2rgb as sk_hsv2rgb

from vlutils.vision.colorSpace import rgb2hsv, hsv2rgb


class TestDataset:
    def testRgb2Hsv(self):
        x = np.random.uniform(low=0.0, high=1.0, size=(1, 3, 255, 255))
        xT = torch.from_numpy(x)
        y = sk_rgb2hsv(x.transpose(0, 2, 3, 1)).transpose(0, 3, 1, 2)
        yT = rgb2hsv(xT)
        assert ((y - yT.numpy()) ** 2).sum() < 1e-12

    def testHsv2Rgb(self):
        x = np.random.uniform(low=0.0, high=1.0, size=(1, 3, 255, 255))
        xT = torch.from_numpy(x)
        y = sk_hsv2rgb(x.transpose(0, 2, 3, 1)).transpose(0, 3, 1, 2)
        yT = hsv2rgb(xT)
        assert ((y - yT.numpy()) ** 2).sum() < 1e-12



"""Module of modified `nn.DataParallel`."""
from itertools import chain
from typing import Callable, Optional, Union, List
import functools
import threading

import torch
from torch import nn
from torch.cuda._utils import _get_device_index
from torch._utils import ExceptionWrapper


__all__ = [
    "parallelFunction",
    "DataParallel"
]


def _get_a_var(obj):
    if isinstance(obj, torch.Tensor):
        return obj

    if isinstance(obj, list) or isinstance(obj, tuple):
        for result in map(_get_a_var, obj):
            if isinstance(result, torch.Tensor):
                return result
    if isinstance(obj, dict):
        for result in map(_get_a_var, obj.items()):
            if isinstance(result, torch.Tensor):
                return result
    return None


def _parallel_apply(modules, funcName: str, inputs, kwargs_tup=None, devices=None):
    """Applies each `module` in :attr:`modules` in parallel on arguments
    contained in :attr:`inputs` (positional) and :attr:`kwargs_tup` (keyword)
    on each of :attr:`devices`.

    Args:
        modules (Module): modules to be parallelized
        inputs (tensor): inputs to the modules
        devices (list of int or torch.device): CUDA devices

    :attr:`modules`, :attr:`inputs`, :attr:`kwargs_tup` (if given), and
    :attr:`devices` (if given) should all have same length. Moreover, each
    element of :attr:`inputs` can either be a single object as the only argument
    to a module, or a collection of positional arguments.
    """
    assert len(modules) == len(inputs)
    if kwargs_tup is not None:
        assert len(modules) == len(kwargs_tup)
    else:
        kwargs_tup = ({},) * len(modules)
    if devices is not None:
        assert len(modules) == len(devices)
    else:
        devices = [None] * len(modules)
    devices = list(map(lambda x: _get_device_index(x, True), devices))
    lock = threading.Lock()
    results = {}
    grad_enabled = torch.is_grad_enabled()

    def _worker(i, module, input, kwargs, device=None):  # pylint: disable = redefined-builtin
        torch.set_grad_enabled(grad_enabled)
        if device is None:
            device = _get_a_var(input).get_device()
        try:
            with torch.cuda.device(device):
                # this also avoids accidental slicing of `input` if it is a Tensor
                if not isinstance(input, (list, tuple)):
                    input = (input,)
                output = getattr(module, funcName)(*input, **kwargs)
            with lock:
                results[i] = output
        except Exception:  # pylint: disable = broad-except
            with lock:
                results[i] = ExceptionWrapper(
                    where="in replica {} on device {}".format(i, device))

    if len(modules) > 1:
        threads = [threading.Thread(target=_worker,
                                    args=(i, module, input, kwargs, device))
                   for i, (module, input, kwargs, device) in
                   enumerate(zip(modules, inputs, kwargs_tup, devices))]

        for thread in threads:
            thread.start()
        for thread in threads:
            thread.join()
    else:
        _worker(0, modules[0], inputs[0], kwargs_tup[0], devices[0])

    outputs = []
    for i in range(len(inputs)):
        output = results[i]
        if isinstance(output, ExceptionWrapper):
            output.reraise()
        outputs.append(output)
    return outputs


def parallelFunction(function: Callable):
    """A decorator for class-methods that want to parallel-apply.

    Example:
    ```python
        class Foo(nn.Module):
            ...

            # This function will be parallel-applied, just like forward(...)
            @parallelFunction
            def add(self, x, y):
                return x + y
    ```
    See also `DataParallel`.

    Args:
        function (Callable): The class-method to be decorated.

    Returns:
        Callable: The decorated class-method.
    """
    @functools.wraps(function)
    def wrappedFn(*args, **kwargs):
        return function(*args, **kwargs)
    return wrappedFn


class DataParallel(nn.DataParallel):
    """Customized `nn.DataParallel`.

    Example:
    ```python
        class Foo(nn.Module):
            ...

            # This function will be parallel-applied, just like forward(...)
            @parallelFunction
            def add(self, x, y):
                return x + y

        foo = Foo()
        foo = DataParallel(foo)
        # parallel-apply
        foo.add(3, 4)
    ```

    Args:
        module (nn.Module): Module to be parallelized.
        device_ids (Optional[List[Union[int, torch.device]]], optional): Devices to be placed on. Defaults to None.
        output_device (Union[int, torch.device], optional): Devices where results placed. Defaults to None.
        dim (int, optional): Which dimension of input to be splitted for parallel. Defaults to 0.
    """
    def __init__(self, module: nn.Module, device_ids: Optional[List[Union[int, torch.device]]] = None, output_device: Union[int, torch.device] = None, dim: int = 0):
        super().__init__(module, device_ids, output_device, dim)
        fns = [getattr(module, fn) for fn in dir(module) if callable(getattr(module, fn))]
        for fn in fns:
            if getattr(fn, "__wrapped__", None) is not None:
                setattr(self, fn.__name__, self._parallelFnWrapper(fn))

    def _parallelFnWrapper(self, func):
        @functools.wraps(func)
        def parallelApply(*inputs, **kwargs):
            if not self.device_ids:
                return func(*inputs, **kwargs)

            for t in chain(self.module.parameters(), self.module.buffers()):
                if t.device != self.src_device_obj:
                    raise RuntimeError("module must have its parameters and buffers "
                                       "on device {} (device_ids[0]) but found one of "
                                       "them on device: {}".format(self.src_device_obj, t.device))

            inputs, kwargs = self.scatter(inputs, kwargs, self.device_ids)
            if len(self.device_ids) == 1:
                return func(*inputs[0], **kwargs[0])
            replicas = self.replicate(self.module, self.device_ids[:len(inputs)])
            outputs = _parallel_apply(replicas, func.__name__, inputs, kwargs, self.device_ids[:len(replicas)])
            return self.gather(outputs, self.output_device)
        return parallelApply



from typing import Union
from pathlib import Path

from PIL import Image


def verifyTruncated(path: Union[str, Path]) -> bool:
    try:
        with Image.open(path) as image:
            image.verify()
            if image.format is None:
                return False
            return True
    except (IOError, OSError):
        return False



from typing import Any, Iterable, List
import abc


class Handler(abc.ABC):
    accumulated: float = 0.0
    length: int = 0

    def __init__(self, format: str = r"%.2f"):
        super().__init__()
        self._format = format

    def to(self, device: Any) -> "Handler":
        return self

    @property
    def ShowInSummary(self) -> bool:
        return True

    @property
    def Length(self) -> int:
        return self.length

    @property
    def Accumulated(self) -> float:
        return self.accumulated

    @property
    def Result(self) -> float:
        return self.accumulated / self.length

    def __call__(self, *args: Any, **kwds: Any):
        results = self.handle(*args, **kwds)
        self.length += len(results)
        self.accumulated += sum(results)

    def __str__(self) -> str:
        return self._format % (self.accumulated / self.length)

    def reset(self):
        self.accumulated = 0.0
        self.length = 0

    @abc.abstractmethod
    def handle(self, *args: Any, **kwds: Any) -> List[float]:
        raise NotImplementedError

class Meters:
    def __init__(self, handlers: Iterable[Handler]):
        self._handlers = list(h for h in handlers)

    def __call__(self, *args, **kwds):
        for handler in self._handlers:
            handler(*args, **kwds)

    def reset(self):
        for handler in self._handlers:
            handler.reset()

    def summary(self, reset: bool = False):
        result = ", ".join("%s: %s" %
            (handler.__class__.__name__, handler)
                for handler
                    in self._handlers if handler.ShowInSummary)
        if reset:
            self.reset()
        return result

    def results(self, reset: bool = False):
        result = {
            handler.__class__.__name__: handler.Result
                for handler
                    in self._handlers
        }
        if reset:
            self.reset()
        return result



import torch


__all__ = [
    "l2DistanceWithNorm",
    "l2Distance"
]


def l2DistanceWithNorm(A: torch.Tensor, B: torch.Tensor):
    diff = ((A.unsqueeze(1) - B) ** 2).sum(2)
    maxi, _ = diff.max(1, keepdim=True)
    norm = diff / maxi
    return norm


def l2Distance(A: torch.Tensor, B: torch.Tensor):
    return ((A.unsqueeze(1) - B) ** 2).sum(2)



from rich.progress import Progress, TaskID, Task


class RichProgress(Progress):
    def get_task(self, taskID: TaskID) -> Task:
        with self._lock:
            return self._tasks[taskID]

    def __enter__(self) -> "RichProgress":
        self.start()
        return self



"""Module for custom nn.Module"""
from typing import Any, Union
from abc import abstractmethod
import types

import torch
from torch import nn


__all__ = [
    "Module"
]


class Module(nn.Module):
    """Custom nn.Module

    Example:
    ```python
        # Custom network
        class Net(Module):
            ...
            # Function used for normal forward
            @Module.register("forward")
            def _ff(self, ...):
                ...
            # Function used for computing loss
            @Module.register("loss")
            def _loss(self, ...):
                ...

        net = Net()
        # Call net._ff
        y = net("forward", x)
        # Call net._loss
        loss = net("loss", y, label)
    ```
    """
    @staticmethod
    def register(key):
        """Decorator for register forward function into module.

        Args:
            key (str): The key for registering a forward function.
        """
        def _wrapped(fn: Any):
            if isinstance(fn, property):
                fn.fget._vlutilsModuleMappedFunction = key
            else:
                fn._vlutilsModuleMappedFunction = key
            return fn
        return _wrapped

    def __init__(self):
        super().__init__()
        self._functions = dict()
        for methodname in dir(self):
            method = getattr(self, methodname)
            if isinstance(method, property):
                method = method.fget
            if hasattr(method, "_vlutilsModuleMappedFunction"):
                self._functions[method._vlutilsModuleMappedFunction] = method

    def _replicate_for_data_parallel(self):
        replica = super()._replicate_for_data_parallel()
        # Redirect mapped function to new replica.
        replica._functions = {k: types.MethodType(v.__func__, replica) for k, v in replica._functions.items()}
        return replica

    @abstractmethod
    def _forward(self, *args: Any, **kwargs: Any) -> Any:
        raise NotImplementedError

    def forward(self, key: Union[torch.Tensor, str], *args: Any, **kwargs: Any) -> Any:
        """Custom forward function

        Args:
            key (str): Key of the target fuction.

        Returns:
            Any: Result.
        """
        if not isinstance(key, str):
            args = (key, ) + args
            return self._forward(*args, **kwargs)
        return self._functions[key](*args, **kwargs)



import logging
from typing import Dict, Callable, Any, List, Tuple, Union

from vlutils.runtime import functionFullName

__all__ = [
    "FrequecyHook"
]


class FrequecyHook:
    """A hook to call function by frequency

    Example:
    ```python
        def foo(*, bar, **_):
            print(bar)

        def lorem(*, ipsum, **_):
            print(ipsum)

        hook = FrequecyHook({
            2: foo,
            6: lorem
        })

        for i in range(1, 10):
            hook(i, bar=f"Hello bar{i}.", ipsum=f"Hello ipsum{i}.")

    >>> Hello bar2.
    Hello bar4.
    Hello bar6.
    Hello ipsum6.
    Hello bar8.
    ```

    Args:
        freqAndHooks (Dict[int, Callable]): The function (`value`) to call every `key` steps.
    """
    def __init__(self, *freqAndHooks: Tuple[int, Callable], logger: Union[logging.Logger, "vlutils.logger.LoggerBase"]=logging):
        self._hooks: Dict[int, List[Callable]] = dict()
        self._logger = logger
        for key, value in freqAndHooks:
            if key not in self._hooks:
                self._hooks[key] = list()
            self._hooks[key].append(value)

    def extend(self, *freqAndHooks: Tuple[int, Callable]):
        for key, value in freqAndHooks:
            if key not in self._hooks:
                self._hooks[key] = list()
            self._hooks[key].append(value)

    def append(self, freq: int, hook: Callable):
        if freq not in self._hooks:
            self._hooks[freq] = list()
        self._hooks[freq].append(hook)

    def remove(self, freq: int):
        self._hooks.pop(freq)

    def __call__(self, step: int, *args: Any, **kwArgs: Any) -> Dict[int, Any]:
        """Check whether the step % key == 0, if True, call value by args and kwArgs.

        Args:
            step (int): Current step.

        Returns:
            Dict[int, Any]: if the function is called, add its key and return value into this dict.
        """
        results = dict()
        for key, value in self._hooks.items():
            if step % key == 0:
                results[key] = list()
                for fn in value:
                    fullName = functionFullName(fn)
                    self._logger.debug("Call %s(...) by FrequecyHook@%d", fullName, key)
                    results[key].append(fn(step, *args, **kwArgs))
        return results

    def __str__(self) -> str:
        pretty = { f"{key}": [f"<{fullName}>" for fullName in (functionFullName(func) for func in value)] for key, value in self._hooks.items() }
        result = ""
        for key, value in pretty.items():
            value = ", ".join(value)
            result += f"    {key}:\r\n        [{value}]\r\n"
        return f"FrequencyHook(\r\n{result})"


class ChainHook:
    def __init__(self, *hooks: Union[Callable, None]) -> None:
        self._hooks: List[Callable] = [h for h in hooks if h is not None]

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        for hook in self._hooks:
            hook(*args, **kwds)

    def __str__(self) -> str:
        hookNames = [functionFullName(h) for h in self._hooks]
        hookNames = ",\r\n    ".join(hookNames)
        return f"ChainHook(\r\n    {hookNames})"



from typing import Any
import re
import yaml
from io import StringIO


def strNoneorEmpty(string: Any):
    return string is None or not (string or string.strip())

def _alignYAML(str, pad=0, aligned_colons=False):
    props = re.findall(r'^\s*[\S]+:', str, re.MULTILINE)
    if not props:
        return str
    longest = max([len(i) for i in props]) + pad
    if aligned_colons:
        return ''.join([i+'\n' for i in map(
                    lambda str: re.sub(r'^(\s*.+?[^:#]): \s*(.*)',
                        lambda m: m.group(1) + ''.ljust(longest-len(m.group(1))-1-pad) + ':'.ljust(pad+1) + m.group(2), str, re.MULTILINE),
                    str.split('\n'))])
    else:
        return ''.join([i+'\n' for i in map(
                    lambda str: re.sub(r'^(\s*.+?[^:#]: )\s*(.*)',
                        lambda m: m.group(1) + ''.ljust(longest-len(m.group(1))+1) + m.group(2), str, re.MULTILINE),
                    str.split('\n'))])

def pPrint(d: dict) -> str:
    """Print dict prettier.

    Args:
        d (dict): The input dict.

    Returns:
        str: Resulting string.
    """
    with StringIO() as stream:
        yaml.safe_dump(d, stream, default_flow_style=False)
        return _alignYAML(stream.getvalue(), pad=1, aligned_colons=True)



"""Module of serialization/deserialization."""
import logging
from logging import Logger
from pathlib import Path
from typing import Any, Dict, Type, TypeVar, get_origin, get_args, _SpecialForm, Union, _GenericAlias
from dataclasses import Field, is_dataclass, asdict
import keyword
from io import StringIO

import yaml

T = TypeVar("T")


__all__ = [
    "read",
    "serialize"
]


def _preprocess(plainStr: str, **varsToReplace):
    for key, value in varsToReplace.items():
        plainStr = plainStr.replace("{{{0}}}".format(key), str(value))
    return plainStr


def _replaceKeyword(parsedYaml: dict) -> dict:
    if not isinstance(parsedYaml, dict):
        return parsedYaml
    newDict = dict()
    for key, value in parsedYaml.items():
        if keyword.iskeyword(key):
            key += "_"
        newDict[key] = _replaceKeyword(value)
    return newDict


def _assert(name, instance, types):
    if not isinstance(instance, types):
        raise TypeError(f"{name} in yaml not match the type definition in {name} ({types}), got {type(instance)}.")


def _serialize(instance: Any) -> dict:
    if isinstance(instance, (str, int, bool, float)):
        return instance
    if isinstance(instance, list, set, tuple):
        return instance.__class__(_serialize(x) for x in instance)
    if isinstance(instance, dict):
        return {k: _serialize(v) for k, v in instance.items()}
    return {k: _serialize(v) for k, v in instance.__dict__}


def _deserialize(parsedYaml: Union[dict, str, int, bool, float, list, set, tuple], classDef: Type[T], logger: Logger) -> T:
    if classDef in (str, int, bool, float):
        return classDef(parsedYaml)
    if classDef in (list, set, tuple):
        return classDef(x for x in parsedYaml)
    if classDef is dict:
        return classDef((k, v) for k, v in parsedYaml.items())
    annotations: Dict[str, Field] = classDef.__dataclass_fields__
    updateDict = dict()
    for attr, fieldDef in annotations.items():
        if attr not in parsedYaml and (fieldDef.default is None):
            raise AttributeError(f"{attr} not found in yaml and no init method provide.")
        if fieldDef.type in (str, int, bool, float):
            _assert(attr, parsedYaml[attr], fieldDef.type)
            updateDict[attr] = parsedYaml[attr]
        elif fieldDef.type in (list, set, tuple):
            _assert(attr, parsedYaml[attr], (list, set, tuple))
            updateDict[attr] = fieldDef.type(_deserialize(x, type(x), logger) for x in parsedYaml[attr])
        elif fieldDef.type is dict:
            _assert(attr, parsedYaml[attr], dict)
            updateDict[attr] = {k: _deserialize(x, type(x), logger) for k, x in parsedYaml[attr].items()}
        else:
            if isinstance(fieldDef.type, _GenericAlias):
                origin = get_origin(fieldDef.type)
                if origin not in (list, set, tuple, dict):
                    raise TypeError(f"{attr} in {classDef} is not any of (str, int, bool, float, list, set, tuple, dict), or generic list/dict/tuple, got {fieldDef.type}.")
                args = get_args(fieldDef.type)
                for arg in args:
                    if issubclass(arg, _SpecialForm):
                        raise NotImplementedError(f"Not support for {attr} which is annotated as a special form {arg}.")
                if origin in (list, set, tuple):
                    _assert(attr, parsedYaml[attr], (list, set, tuple))
                    updateDict[attr] = origin(_deserialize(x, args[0], logger) for x in parsedYaml[attr])
                elif origin is dict:
                    if args[0] is not str:
                        raise TypeError(f"Dict must have str as key, {attr} in {classDef} has {args[0]} as key.")
                    updateDict[attr] = {k: _deserialize(x, args[1], logger) for k, x in parsedYaml[attr].items()}
            elif is_dataclass(fieldDef.type):
                updateDict[attr] = _deserialize(parsedYaml[attr], fieldDef.type, logger)
            else:
                raise TypeError(f"Unrecognized type {fieldDef.type} of {attr} in {fieldDef}.")
    return classDef(**updateDict)


def read(configPath: str, varsToReplace: Union[Dict[str, Any], None], classDef: Type[T], logger: Logger = None) -> T:
    """Read from config.yaml

    Args:
        configPath (str): The path of yaml

    Returns:
        T: The resulting config.
    """
    plainYaml = Path(configPath).read_text()
    if isinstance(varsToReplace, dict):
        plainYaml = _preprocess(plainYaml, **varsToReplace)
    result = _deserialize(_replaceKeyword(yaml.full_load(plainYaml)), classDef, logger or logging)

    # result.summary = types.MethodType(serialize, result)
    return result


def serialize(instance: Any, logger: Logger = None) -> dict:
    """Serialize any object to dict-like

    Args:
        instance (Any): Any object.
        logger (Logger, optional): Logger for logging. Defaults to None.

    Returns:
        dict: The serialized dict.
    """
    if is_dataclass(instance):
        return asdict(instance)
    (logger or logging).debug("Instance is not a dataclass, use custom serialize method.")
    return _serialize(instance)


def summary(instance, logger: Logger = None) -> str:
    """Serialize any object to yaml format summary

    Args:
        instance (Any): Any object.

    Returns:
        str: The serialized string.
    """
    with StringIO() as stream:
        yaml.safe_dump(serialize(instance), stream, default_flow_style=False)
        return stream.getvalue()


class Config:
    @classmethod
    def read(cls, configPath: str, varsToReplace: Dict[str, Any], logger: Logger = None):
        return read(configPath, varsToReplace, type(cls), logger)

    def serialize(self, logger: Logger = None):
        """Serialize self to dict-like

        Args:
            logger (Logger, optional): Logger for logging. Defaults to None.

        Returns:
            dict: The serialized dict.
        """
        return serialize(self, logger)

    def summary(self, logger: Logger = None):
        """Serialize self to yaml format summary

        Args:
            logger (Logger, optional): Logger for logging. Defaults to None.

        Returns:
            str: The serialized string.
        """
        return self.serialize(logger)



"""Package of datasets.

Exports:
    Zip: Wrapper dataset behaves like `zip()`.
    Enumerate: Wrapper dataset behaves like `enumerate()` with element-wise count.
"""
from .wrappers import Zip, Enumerate


__all__ = [
    "Zip",
    "Enumerate"
]



"""Module of wrapper datasets."""
from torch.utils.data import Dataset


__all__ = [
    "Zip",
    "Enumerate"
]


class Zip(Dataset):
    """Wrapper dataset behaves like `zip()`.

    Args:
        *data (List of Dataset): A series of datasets.
    """
    def __init__(self, *datas: Dataset):
        super().__init__()
        self._data = datas

    def __len__(self):
        return len(self._data[0])

    def __getitem__(self, idx):
        return tuple(d[idx] for d in self._data)


class Enumerate(Dataset):
    """Wrapper dataset behaves like `enumerate()` with element-wise count.

    Args:
        data (Dataset): Any datasets.
        mode (str): The representation of the return indices.
                    "asis": -1 -> -1
                    "absolute": -1 -> len - 1
    """
    def __init__(self, data: Dataset, mode: str = "asis"):
        super().__init__()
        self._data = data
        if mode not in ["asis", "absolute"]:
            raise ValueError(f"Given mode not in ['asis', 'absolute'], got {mode}.")
        self._map = {
            "asis": self._asIs,
            "absolute": self._absolute
        }
        self._idxMapping = self._map[mode]

    def mode(self, mode: str = "asis"):
        if mode not in ["asis", "absolute"]:
            raise ValueError(f"Given mode not in ['asis', 'absolute'], got {mode}.")
        self._idxMapping = self._map[mode]

    def _asIs(self, idx):
        return idx

    def _absolute(self, idx):
        return (idx + len(self)) % len(self)

    def __len__(self):
        return len(self._data)

    def __getitem__(self, idx):
        return self._idxMapping(idx), self._data[idx]



"""Module of Registry."""
import functools
import logging
from typing import Callable, Dict, Union, Generic, TypeVar
from vlutils.utils import pPrint


T = TypeVar("T")

__all__ = [
    "Registry"
]


class Registry(Generic[T]):
    """A registry. Inherit from it to create a lots of factories.

    Example:
    ```python
        # Inherit to make a factory.
        class Geometry(Registry):
            ...

        # Register with auto-key "Foo"
        @Geometry.register
        class Foo:
            ...

        # Register with manual-key "Bar"
        @Geometry.register("Bar")
        class Bar:
            ...

        instance = Geometry.get("Foo")()
        assert isinstance(instance, Foo)

        instance = Geometry["Bar"]()
        assert isinstance(instance, Bar)
    ```
    """
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._map: Dict[str, T] = dict()

    @classmethod
    def register(cls, key):
        """Decorator for register anything into registry.

        Args:
            key (str): The key for registering an object.
        """
        if isinstance(key, str):
            def insert(value):
                cls._map[key] = value
                return value
            return insert
        else:
            cls._map[key.__name__] = key
            return key

    @classmethod
    def get(cls, key: str, logger: Union[logging.Logger, "vlutils.logger.LoggerBase"] = logging.root) -> T:
        """Get an object from registry.

        Args:
            key (str): The key for the registered object.
        """
        result = cls._map[key]
        if isinstance(result, functools.partial):
            logger.debug("Get <%s.%s> from \"%s\".", result.func.__module__, result.func.__qualname__, cls.__name__)
        else:
            logger.debug("Get <%s.%s> from \"%s\".", result.__module__, result.__qualname__, cls.__name__)
        return result

    @classmethod
    def summary(cls) -> str:
        """Get registry summary.
        """
        return pPrint({
            k: v.__module__ + v.__name__ for k, v in cls._map.items()
        })



import pytest

from torch import nn
import torch

from vlutils.base import DataParallel, Module, Registrar, Restorable
from vlutils.metrics.helpers import recursiveCompare


def _initToOnes(m: nn.Module):
    if isinstance(m, (nn.Linear, nn.Conv2d)):
        nn.init.ones_(m.weight)
        if m.bias is not None:
            nn.init.ones_(m.bias)


class Factory(Registrar):
    ...


@Factory.register
@Factory.register("AnotherKey")
class ForFactory:
    ...


class NewFactory(Factory):
    ...


@NewFactory.register("ForNewFactory")
@NewFactory.register("ForNewFactory1")
class ForNewFactory:
    ...


class NestedRestorable(Restorable):
    def __init__(self):
        super().__init__()
        self._t = torch.ones([3])


class SimpleRestorable(Restorable):
    def __init__(self):
        super().__init__()
        self._a = 3
        self._b = "str"
        self._c = NestedRestorable()


class TestBase:
    @pytest.fixture
    def getModule(self):
        class linear(Module):
            def __init__(self, d1in: int, d1out: int, d2in: int, d2out: int, bias: bool = True):
                super().__init__()
                self._linear1 = nn.Linear(d1in, d1out, bias)
                self._linear2 = nn.Linear(d2in, d2out, bias)
                self.apply(_initToOnes)

            def _forward(self, x1, x2):
                return self._linear1(x1), self._linear2(x2)

            @Module.register("_ff1")
            def _ff1(self, x):
                return self._linear1(x)

            @Module.register("_ff2")
            def _ff2(self, x):
                return self._linear2(x)

        return linear(3, 6, 7, 4)

    @pytest.fixture
    def getLegacyModule(self):
        class linear(nn.Module):
            def __init__(self, din: int, dout: int, bias: bool = True):
                super().__init__()
                self._linear = nn.Linear(din, dout, bias)
                self.apply(_initToOnes)

            def forward(self, x):
                return self._linear(x)
        return linear(3, 6), linear(7, 4)

    @pytest.fixture
    def getDataParallel(self, getLegacyModule):
        return DataParallel(getLegacyModule[0].cuda())

    @pytest.fixture
    def getLegacyDataParallel(self, getLegacyModule):
        return nn.DataParallel(getLegacyModule[0].cuda())

    @pytest.fixture
    def getInput(self):
        return torch.randn(100, 3), torch.randn(45, 7)

    @pytest.mark.slow
    @pytest.mark.skipif(not torch.cuda.is_available(), reason="System not CUDA-enabled.")
    def testDataParallel(self, getInput, getDataParallel, getLegacyDataParallel):
        x = getInput[0].cuda()
        # Basic usage
        y1 = getDataParallel(x)
        y2 = getDataParallel(x)
        assert y1.shape == y2.shape
        assert nn.functional.mse_loss(y1, y2) < 1e-12

    @pytest.mark.slow
    def testModule(self, getInput, getModule, getLegacyModule):
        x1, x2 = getInput
        legacyModule1, legacyModule2 = getLegacyModule
        y1 = legacyModule1(x1)
        y2 = legacyModule2(x2)
        # Basic forward
        y11, y12 = getModule(x1, x2)
        # Mapped forward
        y21, y22 = getModule("_ff1", x1), getModule("_ff2", x2)
        assert nn.functional.mse_loss(y1, y11) < 1e-12
        assert nn.functional.mse_loss(y1, y21) < 1e-12
        assert nn.functional.mse_loss(y2, y12) < 1e-12
        assert nn.functional.mse_loss(y2, y22) < 1e-12

    def testFactory(self):
        # Mapping independency
        assert "ForFactory" not in NewFactory._map
        assert "ForNewFactory" not in Factory._map
        # Basic function
        assert ForFactory == Factory.get(ForFactory.__name__) == Factory.get("AnotherKey") == Factory["ForFactory"]
        assert ForNewFactory == NewFactory.get("ForNewFactory") == NewFactory.get("ForNewFactory1") == NewFactory["ForNewFactory"]

    def testRestorable(self):
        instance = SimpleRestorable()
        assert recursiveCompare(instance.state_dict(), {
            "_a": 3,
            "_b": "str",
            "_c": {
                "_t": torch.ones([3])
            }
        })
        instance.load_state_dict({
            "_a": "mm",
            "_b": 6,
            "_c": {
                "_t": torch.zeros([4])
            }
        })
        assert recursiveCompare(instance.state_dict(), {
            "_a": "mm",
            "_b": 6,
            "_c": {
                "_t": torch.zeros([4])
            }
        })



import numpy as np
import pytest

from vlutils.datasets import Zip, Enumerate


class TestDataset:
    @pytest.fixture
    def getData(self):
        return np.arange(0, 5), np.arange(1, 6), np.arange(0, 10).reshape((-1, 2))

    def testZip(self, getData):
        a, b, c = getData
        zipDataset = Zip(a, b, c)
        assert len(zipDataset[0]) == 3
        aa, bb, cc = zipDataset[-1]
        assert aa == 4 and bb == 5 and np.array_equal(np.array([8, 9]), cc)
        assert len(zipDataset) == 5

    def testEnumerate(self, getData):
        a, b, c = getData
        zipDataset = Zip(a, b, c)
        enumDataset = Enumerate(zipDataset, "asis")
        assert len(enumDataset) == 5
        i, (aa, bb, cc) = enumDataset[-1]
        assert aa == 4 and bb == 5 and np.array_equal(np.array([8, 9]), cc) and i == -1

        enumDataset = Enumerate(zipDataset, "absolute")
        assert len(enumDataset) == 5
        i, (aa, bb, cc) = enumDataset[-1]
        assert aa == 4 and bb == 5 and np.array_equal(np.array([8, 9]), cc) and i == 4
        i, (aa, bb, cc) = enumDataset[2]
        assert aa == 2 and bb == 3 and np.array_equal(np.array([4, 5]), cc) and i == 2



import torch


__all__ = [
    "rgb2hsv",
    "hsv2rgb"
]


def rgb2hsv(x: torch.Tensor, mode: str = "normal", eps: float = 1e-8) -> torch.Tensor:
    """Map rgb color space to HSV color space

    before call this function, rgb should first reshape to [N, C, H, W] and normalize to [0, 1]

    Args:
        x (torch.Tensor): [N, C, H, W] Image tensor.
        mode (str): Routine to comute hsv, one of "normal", "advance".
        eps (float, optional): Epsilon for small values. Defaults to 1e-8.

    returns:
        torch.Tensor: [N, C, H, W] Converted tensor.
    """
    return {
        "normal": _hsvNormal,
        "advance": _hsvAdvance
    }[mode](x, eps)


def _hsvNormal(x: torch.Tensor, eps: float) -> torch.Tensor:
    cMax, maxIdx = x.max(1)
    cMin, minIdx = x.min(1)

    sub = cMax - cMin

    # don't need requires_grad
    hsv = torch.empty_like(x)

    idx = maxIdx == 0
    hsv[:, 0][idx] = (x[:, 1][idx] - x[:, 2][idx]) / sub[idx]

    idx = maxIdx == 1
    hsv[:, 0][idx] = 2. + (x[:, 2][idx] - x[:, 0][idx]) / sub[idx]

    idx = maxIdx == 2
    hsv[:, 0][idx] = 4. + (x[:, 0][idx] - x[:, 1][idx]) / sub[idx]

    hsv[:, 0] = hsv[:, 0] / 6. % 1.
    hsv[:, 0][maxIdx == minIdx] = 0.0

    s = 1 - cMin / cMax
    s[cMax.abs() < eps] = 0.0

    hsv[:, 1] = s
    hsv[:, 2] = cMax

    return hsv


def _hsvAdvance(x: torch.Tensor, eps: float) -> torch.Tensor:
    raise NotImplementedError


def hsv2rgb(x: torch.Tensor, mode: str = "normal", eps: float = 1e-8) -> torch.Tensor:
    """Map rgb color space to HSV color space

    before call this function, rgb should first reshape to [N, C, H, W] and normalize to [0, 1]

    Args:
        x (torch.Tensor): [N, C, H, W] Image tensor.
        mode (str): Routine to comute hsv, one of "normal", "advance".
        eps (float, optional): Epsilon for small values. Defaults to 1e-8.

    returns:
        torch.Tensor: [N, C, H, W] Converted tensor.
    """
    return {
        "normal": _rgbNormal,
        "advance": _rgbAdvance
    }[mode](x, eps)


def _rgbNormal(x: torch.Tensor, eps: float) -> torch.Tensor:
    hi = (x[:, 0] * 6).floor()
    f = x[:, 0] * 6 - hi
    p = x[:, 2] * (1 - x[:, 1])
    q = x[:, 2] * (1 - f * x[:, 1])
    t = x[:, 2] * (1 - (1 - f) * x[:, 1])
    v = x[:, 2]

    hi = torch.stack([hi, hi, hi], axis=1).byte() % 6
    return ((hi == 0) * torch.stack((v, t, p), axis=1)) + ((hi == 1) * torch.stack((q, v, p), axis=1)) + ((hi == 2) * torch.stack((p, v, t), axis=1)) + ((hi == 3) * torch.stack((p, q, v), axis=1)) + ((hi == 4) * torch.stack((t, p, v), axis=1)) + ((hi == 5) * torch.stack((v, p, q), axis=1))


def _rgbAdvance(x: torch.Tensor, eps: float) -> torch.Tensor:
    raise NotImplementedError
